from hotsos.core.plugins.openstack.common import OpenStackChecks
from hotsos.core.plugins.openstack.neutron import NeutronHAInfo
from hotsos.core.plugintools import summary_entry


class OpenStackSummary(OpenStackChecks):
    """ Implementation of OpenStack summary. """
    summary_part_index = 0

    @summary_entry('release', 0)
    def summary_release(self):
        return {'name': self.release_name,
                'days-to-eol': self.days_to_eol}

    @summary_entry('services', 1)
    def summary_services(self):
        """Get string info for running services."""
        if self.systemd.services:
            return self.systemd.summary
        if self.pebble.services:
            return self.pebble.summary

        return None

    @summary_entry('dpkg', 2)
    def summary_dpkg(self):
        # require at least one core package to be installed to include
        # this in the report.
        if self.apt.core:
            return self.apt.all_formatted

        return None

    @summary_entry('docker-images', 3)
    def summary_docker_images(self):
        # require at least one core image to be in-use to include
        # this in the report.
        if self.docker.core:
            return self.docker.all_formatted

        return None

    @staticmethod
    @summary_entry('neutron-l3ha', 4)
    def summary_neutron_l3ha():
        routers = {}
        ha_info = NeutronHAInfo()
        for router in ha_info.ha_routers:
            state = router.ha_state
            if state in routers:
                routers[state] += 1
            else:
                routers[state] = 1

        return routers or None



from .common import CephChecks, CephConfig, CephDaemonAllOSDsFactory
from .cluster import CephCluster, CephCrushMap

__all__ = [CephChecks.__name__,
           CephConfig.__name__,
           CephDaemonAllOSDsFactory.__name__,
           CephCluster.__name__,
           CephCrushMap.__name__]



from hotsos.core.plugins.lxd import LXD, LXDChecks
from hotsos.core.plugintools import summary_entry


class LXDSummary(LXDChecks):
    """ Implementation of LXD summary. """
    summary_part_index = 0

    @summary_entry('services', 0)
    def summary_services(self):
        if self.systemd.services:
            return self.systemd.summary

        return None

    @summary_entry('snaps', 1)
    def summary_snaps(self):
        if self.snaps:
            return self.snaps.all_formatted

        return None

    @summary_entry('dpkg', 2)
    def summary_dpkg(self):
        if self.apt:
            return self.apt.all_formatted

        return None

    @staticmethod
    @summary_entry('instances', 3)
    def summary_instances():
        return LXD().instances or None



import glob
import os
import re
from datetime import datetime, timezone
from functools import cached_property

import dateutil
import pytz
from dateutil import parser as dateutil_parser
# NOTE: we import direct from searchkit rather than hotsos.core.search to
#       avoid circular dependency issues.
from searchkit import (
    FileSearcher,
    SearchDef,
    SequenceSearchDef,
)
from hotsos.core.config import HotSOSConfig
from hotsos.core.factory import FactoryBase
from hotsos.core.host_helpers import CLIHelper, CLIHelperFile
from hotsos.core.host_helpers.common import ServiceManagerBase
from hotsos.core.log import log
from hotsos.core.utils import sorted_dict


class SystemdService():
    """ Representation of systemd service. """
    def __init__(self, name, state, has_instances=False):
        self.name = name
        self.state = state
        self.has_instances = has_instances

    @cached_property
    def _tzinfos(self):
        """ Generates timezone name to zone mappings for common timezones.

        See https://dateutil.readthedocs.io/en/stable/parser.html#functions for
        how it is consumed. dateutil needs timezone abbreviated to timezone
        full name mappings in order to be able to parse timestamps with tz
        abbreviations.

        @return: dictionary of name: timezone mappings.
        """
        def fetch():
            for zone in pytz.common_timezones:
                try:
                    tzdate = pytz.timezone(zone).localize(datetime.utcnow(),
                                                          is_dst=None)
                except pytz.NonExistentTimeError:
                    pass
                else:
                    tzinfo = dateutil.tz.gettz(zone)
                    if tzinfo:
                        yield tzdate.tzname(), tzinfo

        return dict(fetch())

    @cached_property
    def start_time(self):
        """ Get most recent start time of this service unit.

        We first look in systemd journal since that should be the fastest and
        if not found (perhaps because service not restarted for a long time)
        we look in service status.

        @returns: datetime.datetime object or None if time not found.
        """
        log.debug("fetching start time for svc %s", self.name)
        cli = CLIHelper()
        # must be in short-iso format e.g. 2023-07-04T00:05:23+0100
        cexpr = re.compile(r"^(\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\+\d{4})"
                           r"\s+.+: (Started|Starting)")
        journal = cli.journalctl(unit=self.name)
        last = None
        for line in journal:
            ret = cexpr.search(line)
            if ret:
                last = ret.group(1)

        if last:
            return datetime.strptime(last, "%Y-%m-%dT%H:%M:%S%z")

        log.debug("start time not found in journal, trying service status")
        # NOTE: should consider getting service status directly rather than
        #       searching in all but currently do this to have parity with
        #       sosreport.
        fs = FileSearcher(decode_errors='backslashreplace')
        # The following expressions need to take account of control characters
        # that might exist in the output e.g. line can start with '*' or U+25CF
        # Active: active (running) since Wed 2022-02-09 22:38:17 UTC; 17h ago
        seqdef = SequenceSearchDef(
                    start=SearchDef(rf'\S+ ({self.name}.service) -'),
                    body=SearchDef(r"\s+Active: active \(?\S*\)?\s*since "
                                   r"\S{3} (\d{4}-\d{2}-\d{2} "
                                   r"\d{2}:\d{2}:\d{2} [\w\+:-]+);"),
                    end=SearchDef(r'(\S+) \S+.service'),
                    tag='systemd')
        with CLIHelperFile() as cli:
            fs.add(seqdef, path=cli.systemctl_status_all())
            sections = list(fs.run().find_sequence_sections(seqdef).values())
            if len(sections) == 0:
                log.warning("no active status found for %s.service (state=%s)",
                            self.name, self.state)
                return None

            if len(sections) > 1:
                log.warning("more than one status found for %s.service",
                            self.name)

            for result in sections[0]:
                if result.tag == seqdef.body_tag:
                    return dateutil_parser.parse(result.get(1),
                                                 tzinfos=self._tzinfos)
        log.debug("no start time identified for svc %s (state=%s)", self.name,
                  self.state)
        return None

    @cached_property
    def start_time_secs(self):
        """ Get most recent start time of this service unit in seconds.

        @returns: posix timestamp
        """
        t = self.start_time
        if t is None:
            return 0

        if t.utcoffset() is None:
            t = t.replace(tzinfo=timezone.utc)
        else:
            t = t.astimezone(tz=timezone.utc)

        return t.timestamp()

    @property
    def memory_current_kb(self):
        """ Returns service current memory usage in kbytes.

        See https://www.kernel.org/doc/Documentation/cgroup-v1/memory.txt
        See https://www.kernel.org/doc/Documentation/cgroup-v2.txt
        """
        cgroupv1 = os.path.join(HotSOSConfig.data_root, 'sys/fs/cgroup',
                                f"memory/system.slice/{self.name}.service",
                                'memory.stat')
        cgroupv2 = os.path.join(HotSOSConfig.data_root, 'sys/fs/cgroup',
                                'system.slice', f"{self.name}.service",
                                'memory.current')
        if os.path.exists(cgroupv1):
            total_usage = {}
            fs = FileSearcher()
            fs.add(SearchDef(r'(cache|rss|swap) (\d+)'), path=cgroupv1)
            for result in fs.run().get(cgroupv1, {}):
                total_usage[result.get(1)] = int(result.get(2))

            if len(total_usage) == 0:
                log.warning("failed to identify mem usage info for %s in %s",
                            f"{self.name}.service", cgroupv1)

            total = sum(total_usage.values())
            if total == 0:
                return total

            return int(total / 1024)

        # NOTE: memory.current (v2) is assumed not to be an approximation like
        #       memory.usage_in_bytes in v1
        if not os.path.exists(cgroupv2):
            log.warning("service memory info not found at %s", cgroupv2)
            return 0

        with open(cgroupv2, encoding='utf-8') as fd:
            total = int(fd.read())
            if total == 0:
                return total

            return int(total / 1024)

    def __repr__(self):
        return (f"name={self.name}, state={self.state}, "
                f"start_time={self.start_time}, "
                f"has_instances={self.has_instances}")


class SystemdHelper(ServiceManagerBase):
    """ Helper class used to query systemd services. """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._cached_unit_files_exprs = {}

    @cached_property
    def _systemctl_list_units(self):
        return CLIHelper().systemctl_list_units()

    @cached_property
    def _ps(self):
        return CLIHelper().ps()

    def _has_unit_instances(self, expr, ignore_exited=True):
        """
        Determine if unit has instances that match the given expression.

        @param expr: expression used to match one or more units in
                     systemctl list-units.
        @param ignore_exited: filter units that are "exited"
        """
        for line in self._systemctl_list_units:
            ret = re.compile(expr).match(line)
            if ret:
                if ignore_exited and ret.group(4) == 'exited':
                    continue

                return True

        return False

    def _unit_files_expr(self, svc_name_expr):
        """
        Returns search expression used to match unit files based in service
        name expression.

        @param svc_name_expr: expression to match service name.
        """
        if svc_name_expr in self._cached_unit_files_exprs:
            return self._cached_unit_files_exprs[svc_name_expr]

        # Add snap prefix/suffixes
        base_expr = rf"(?:snap\.)?{svc_name_expr}(?:\.daemon)?"
        # NOTE: we include indirect services (ending with @) so that
        #       we can search for related units later.
        unit_expr = rf'^\s*({base_expr}@?)\.service\s+(\S+)'
        # match entries in systemctl list-unit-files
        self._cached_unit_files_exprs[svc_name_expr] = re.compile(unit_expr)
        return self._cached_unit_files_exprs[svc_name_expr]

    @cached_property
    def services(self):
        """
        Return a dict of identified systemd services and their state.

        Service units are either direct or indirect. We unify these types,
        taking the state of whichever is actually in use i.e. has in-memory
        instances. Enabled units are aggregated but masked units are not so
        that they can be identified and reported.
        """
        _services = {}
        for line in CLIHelper().systemctl_list_unit_files():
            for svc_name_expr in self._service_exprs:
                # check each matched unit file for instances
                ret = self._unit_files_expr(svc_name_expr).match(line)
                if not ret:
                    continue

                unit = ret.group(1)
                state = ret.group(2)
                has_instances = False
                units_expr = r"\*?\s+({})\.service\s+(\S+)\s+(\S+)\s+(\S+)"
                if unit.endswith('@'):
                    # indirect or "template" units can have "instantiated"
                    # units where only the latter represents whether the
                    # unit is in use. If an indirect unit has instantiated
                    # units we use them to represent the state of the
                    # service.
                    units_expr = units_expr.format(unit + r'\S+')
                    unit = unit.partition('@')[0]
                else:
                    units_expr = units_expr.format(unit)

                if state != 'disabled':
                    if self._has_unit_instances(units_expr):
                        has_instances = True
                elif unit in _services:
                    # don't override enabled with disabled
                    continue

                if state == 'indirect' and not has_instances:
                    continue

                if unit in _services:
                    if (_services[unit].state == 'indirect' and
                            _services[unit].has_instances):
                        continue

                _services[unit] = SystemdService(unit, state, has_instances)

        # NOTE: assumes that indirect instances always supersede direct ones
        #       i.e. you can't have both.
        for info in _services.values():
            if info.state == 'indirect':
                info.state = 'enabled'

        return _services

    @property
    def masked_services(self):
        """ Returns a list of masked services. """
        if not self.services:
            return []

        return sorted(self._service_info.get('masked', []))

    def get_services_expanded(self, name):
        _expanded = []
        for line in self._systemctl_list_units:
            expr = rf'^\s*({name}(@\S*)?)\.service'
            ret = re.compile(expr).match(line)
            if ret:
                _expanded.append(ret.group(1))

        if not _expanded:
            _expanded = [name]

        return _expanded

    @cached_property
    def _service_filtered_ps(self):
        """
        For each service get list of processes started by that service and
        get their corresponding binary name from ps.

        Returns list of lines from ps that match the service pids.
        """
        ps_filtered = []
        path = os.path.join(HotSOSConfig.data_root,
                            'sys/fs/cgroup/unified/system.slice')
        if not os.path.exists(path):
            # Seems that this path changed between Ubuntu Focal and Jammy
            # releases.
            path = os.path.join(HotSOSConfig.data_root,
                                'sys/fs/cgroup/system.slice')

        for svc in self.services:
            for svc in self.get_services_expanded(svc):
                _path = os.path.join(path, f"{svc}.service",
                                     'cgroup.procs')
                if not os.path.exists(_path):
                    _path = glob.glob(os.path.join(path, 'system-*.slice',
                                                   f"{svc}.service",
                                                   'cgroup.procs'))
                    if not _path or not os.path.exists(_path[0]):
                        continue

                    _path = _path[0]

                with open(_path, encoding='utf-8') as fd:
                    for pid in fd:
                        for line in self._ps:
                            if re.match(rf'^\S+\s+{int(pid)}\s+', line):
                                ps_filtered.append(line)
                                break

        return ps_filtered

    @cached_property
    def processes(self):
        """
        Identify running processes from ps that are associated with resolved
        systemd services. The search pattern used to identify a service is also
        used to match the process binaryc/cmd name.

        Accounts for different types of process cmd path e.g.

        /snap/<name>/1830/<svc>
        /usr/bin/<svc>

        and filter e.g.

        /var/lib/<svc> and /var/log/<svc>

        Returns a dictionary of process names along with the number of each.
        """
        _proc_info = {}
        for line in self._service_filtered_ps:
            for expr in self._service_exprs:
                cmd = self.get_cmd_from_ps_line(line, expr)
                if not cmd:
                    continue

                if cmd in _proc_info:
                    _proc_info[cmd] += 1
                else:
                    _proc_info[cmd] = 1

        return _proc_info

    @property
    def _service_info(self):
        """Return a dictionary of systemd services grouped by state. """
        info = {}
        for svc, obj in sorted_dict(self.services).items():
            state = obj.state
            if state not in info:
                info[state] = []

            info[state].append(svc)

        return info

    @property
    def _process_info(self):
        """Return a list of processes associated with services. """
        return [f"{name} ({count})"
                for name, count in sorted_dict(self.processes).items()]

    @property
    def summary(self):
        """
        Output a dict summary of this class i.e. services, their state and any
        processes run by them.
        """
        return {'systemd': self._service_info,
                'ps': self._process_info}


class ServiceFactory(FactoryBase):
    """
    Factory to dynamically create SystemdService objects for given services.

    Service objects are returned when a getattr() is done on this object using
    the name of the service as the attr name.
    """

    def __getattr__(self, svc):
        log.debug("creating service object for %s", svc)
        return SystemdHelper([svc]).services.get(svc)



from hotsos.core.plugins.maas import MAASChecks
from hotsos.core.plugintools import summary_entry


class MAASSummary(MAASChecks):
    """ Implementation of MAAS summary. """
    summary_part_index = 0

    @summary_entry('services', 0)
    def summary_services(self):
        if self.systemd.services:
            return self.systemd.summary

        return None

    @summary_entry('dpkg', 1)
    def summary_dpkg(self):
        if self.apt.core:
            return self.apt.all_formatted

        return None

    @summary_entry('snaps', 2)
    def summary_snaps(self):
        if self.snaps.core:
            return self.snaps.all_formatted

        return None



from hotsos.core.plugins.storage.ceph.common import CephChecks
from hotsos.core.utils import sorted_dict
from hotsos.core.plugintools import summary_entry


class CephSummary(CephChecks):
    """ Implementation of Ceph summary. """
    summary_part_index = 0

    @summary_entry('release', 0)
    def summary_release(self):
        return {'name': self.release_name,
                'days-to-eol': self.days_to_eol}

    @summary_entry('services', 1)
    def summary_services(self):
        """Get string info for running services."""
        if self.systemd.services:
            return self.systemd.summary

        if self.pebble.services:
            return self.pebble.summary

        return None

    @summary_entry('dpkg', 2)
    def summary_dpkg(self):
        # require at least one core package to be installed to include
        # this in the report.
        if self.apt.core:
            return self.apt.all_formatted

        return None

    @summary_entry('snaps', 2)
    def summary_snaps(self):
        if self.snaps.core:
            return self.snaps.all_formatted

        return None

    @summary_entry('status', 3)
    def summary_status(self):
        return self.cluster.health_status or None

    @summary_entry('network', 4)
    def summary_network(self):
        """ Identify ports used by Ceph daemons, include them in output
        for informational purposes.
        """
        net_info = {}
        for config, port in self.bind_interfaces.items():
            net_info[config] = port.to_dict()

        if net_info:
            return net_info

        return None

    @summary_entry('osd-pgs-near-limit', 5)
    def summary_osd_pgs_near_limit(self):
        if self.cluster.osds_pgs_above_max:
            return self.cluster.osds_pgs_above_max

        return None

    @summary_entry('osd-pgs-suboptimal', 6)
    def summary_osd_pgs_suboptimal(self):
        if self.cluster.osds_pgs_suboptimal:
            return self.cluster.osds_pgs_suboptimal

        return None

    @summary_entry('versions', 7)
    def summary_versions(self):
        versions = self.cluster.ceph_daemon_versions_unique()
        return versions or None

    @summary_entry('mgr-modules', 8)
    def summary_mgr_modules(self):
        return self.cluster.mgr_modules or None

    @summary_entry('local-osds', 9)
    def summary_local_osds(self):
        if self.local_osds:
            osds = {}
            for osd in self.local_osds:
                osds.update(osd.to_dict())

            return sorted_dict(osds)

        return None

    @summary_entry('crush-rules', 10)
    def summary_crush_rules(self):
        return self.cluster.crush_map.rules or None

    @summary_entry('large-omap-pgs', 11)
    def summary_large_omap_pgs(self):
        return self.cluster.large_omap_pgs or None



from hotsos.core.plugins.sosreport import SOSReportChecks
from hotsos.core.plugintools import summary_entry


class SOSReportSummary(SOSReportChecks):
    """ Implementation of SOSReport summary. """
    summary_part_index = 0

    @summary_entry('version', 0)
    def summary_version(self):
        if self.version is not None:
            return self.version

        return None

    @summary_entry('dpkg', 1)
    def summary_dpkg(self):
        if self.apt.core:
            return self.apt.all_formatted

        return None

    @summary_entry('plugin-timeouts', 2)
    def summary_plugin_timeouts(self):
        if self.timed_out_plugins:
            return self.timed_out_plugins

        return None



from hotsos.core.plugins.pacemaker import PacemakerChecks
from hotsos.core.plugintools import summary_entry


class PacemakerSummary(PacemakerChecks):
    """ Implementation of Pacemaker summary. """
    summary_part_index = 0

    @summary_entry('services', 0)
    def summary_services(self):
        if self.systemd.services:
            return self.systemd.summary

        return None

    @summary_entry('dpkg', 1)
    def summary_dpkg(self):
        return self.apt.all_formatted or None

    @summary_entry('nodes', 2)
    def summary_nodes(self):
        nodes = {}
        if self.online_nodes:
            nodes["online"] = self.online_nodes
        if self.offline_nodes:
            nodes["offline"] = self.offline_nodes

        return nodes or None



from hotsos.core.plugins.kubernetes import KubernetesChecks
from hotsos.core.plugintools import summary_entry


class KubernetesSummary(KubernetesChecks):
    """ Implementation of Kubernetes summary. """
    summary_part_index = 0

    @summary_entry('services', 1)
    def summary_services(self):
        if self.systemd.services:
            return self.systemd.summary
        if self.pebble.services:
            return self.pebble.summary

        return None

    @summary_entry('snaps', 2)
    def summary_snaps(self):
        return self.snaps.all_formatted or None

    @summary_entry('dpkg', 3)
    def summary_dpkg(self):
        return self.apt.all_formatted or None

    @summary_entry('pods', 4)
    def summary_pods(self):
        return self.pods or None

    @summary_entry('containers', 5)
    def summary_containers(self):
        return self.containers or None

    @summary_entry('flannel', 6)
    def summary_flannel(self):
        info = {}
        for port in self.flannel_ports:
            info[port.name] = port.encap_info
            if port.addresses:
                info[port.name]['addr'] = port.addresses[0]

        return info or None



from hotsos.core.plugins.mysql import MySQLChecks
from hotsos.core.plugintools import summary_entry


class MySQLSummary(MySQLChecks):
    """ Implementation of MySQL summary. """
    summary_part_index = 0

    @summary_entry('services', 0)
    def summary_services(self):
        if self.systemd.services:
            return self.systemd.summary
        if self.pebble.services:
            return self.pebble.summary

        return None

    @summary_entry('dpkg', 1)
    def summary_dpkg(self):
        if self.apt_info.core:
            return self.apt_info.all_formatted

        return None



import abc
import copy
from collections import UserDict

from hotsos.core.exceptions import (
    NameAlreadyRegisteredError,
)


class ConfigOpt():
    """ Basic information required to define a config option. """
    def __init__(self, name, description, default_value, value_type):
        self.name = name
        self.description = description
        self.default_value = default_value
        self.value_type = value_type


class ConfigOptGroupBase(UserDict):
    """ Base class for defining groups of config options. """
    def __init__(self):
        super().__init__()
        self.opts = {}

    @property
    @abc.abstractmethod
    def name(self):
        """ OptGroup name """

    def __getattribute__(self, name):
        """ Ensure dict contains all registered options. """
        if name == 'data':
            return {_name: opt.default_value
                    for _name, opt in self.opts.items()}

        return super().__getattribute__(name)

    def add(self, opt):
        self.opts[opt.name] = opt


class HotSOSConfigOpts(ConfigOptGroupBase):
    """ Group of hotsos common options. """
    def __init__(self):
        super().__init__()
        self.add(ConfigOpt(name='data_root',
                           description=('This is the filesystem root used for '
                                        'all files and is typically / or a '
                                        'path to a sosreport'),
                           default_value=None, value_type=str))
        self.add(ConfigOpt(name='force_mode',
                           description=('Plugin execution is usually gated '
                                        'on a set of conditions. Setting this '
                                        'to True bypasses these conditions '
                                        'and forces all plugins to run'),
                           default_value=False, value_type=bool))
        self.add(ConfigOpt(name='global_tmp_dir',
                           description=('A temporary directory created at the '
                                        'start of execution and visible to '
                                        'all plugins.'),
                           default_value=None, value_type=str))
        self.add(ConfigOpt(name='plugin_tmp_dir',
                           description=('A temporary directory created for '
                                        'each plugin.'),
                           default_value=None, value_type=str))
        self.add(ConfigOpt(name='use_all_logs',
                           description=('Automatically convert log paths to '
                                        'glob e.g. <path> becomes <path>*'),
                           default_value=False, value_type=bool))
        self.add(ConfigOpt(name='machine_readable',
                           description=('If set to True, the summary output'
                                        'will contain extra information '
                                        'that might be useful if the output '
                                        'is being read by an application.'),
                           default_value=False, value_type=bool))
        self.add(ConfigOpt(name='part_name',
                           description=('Plugins have many parts each with '
                                        'its own name and this will be set to '
                                        'the current part being executed.'),
                           default_value=None, value_type=str))
        self.add(ConfigOpt(name='repo_info',
                           description=('Source repository sha1 from which '
                                        'the current build was created'),
                           default_value=None, value_type=str))
        self.add(ConfigOpt(name='hotsos_version',
                           description='Version of hotsos being run.',
                           default_value=None, value_type=str))
        self.add(ConfigOpt(name='debug_mode',
                           description='Set to True to enable debug logging '
                                       'to standard output',
                           default_value=False, value_type=bool))
        self.add(ConfigOpt(name='plugin_yaml_defs',
                           description='Path to yaml-defined checks.',
                           default_value=None, value_type=str))
        self.add(ConfigOpt(name='templates_path',
                           description='Path to jinja templates.',
                           default_value='templates', value_type=str))
        self.add(ConfigOpt(name='plugin_name',
                           description='Name of current plugin being executed',
                           default_value=None, value_type=str))
        self.add(ConfigOpt(name='event_tally_granularity',
                           description=("By default event tallies are listed "
                                        "by date in the summary. This option "
                                        "can be set to one of 'date' or "
                                        "'time' to get the corresponding "
                                        "granularity of results."),
                           default_value='date', value_type=str))
        self.add(ConfigOpt(name='command_timeout',
                           description=("Maximum time in seconds before "
                                        "command execution will timeout. Used "
                                        "by host_helpers.cli when executing "
                                        "binary commands"),
                           default_value=300, value_type=int))
        self.add(ConfigOpt(name='event_filter',
                           description=("Name of event to run. Set this if "
                                        "you want to run a single event. "
                                        "Useful for testing/debugging"),
                           default_value='', value_type=str))
        self.add(ConfigOpt(name='scenario_filter',
                           description=("Name of scenario to run. Set this if "
                                        "you want to run a single scenario. "
                                        "Useful for testing/debugging"),
                           default_value='', value_type=str))
        self.add(ConfigOpt(name='debug_log_levels',
                           description=("Debug mode log levels for "
                                        "submodules/dependencies"),
                           default_value={'propertree': 'WARNING',
                                          'searchkit': 'DEBUG'},
                           value_type=dict))

    @property
    def name(self):
        return 'hotsos'


class SearchtoolsConfigOpts(ConfigOptGroupBase):
    """ Group of search options. """
    def __init__(self):
        super().__init__()
        self.add(ConfigOpt(name='max_parallel_tasks',
                           description=('Maximum parallelism for searching '
                                        'files concurrently'),
                           default_value=8, value_type=int))
        self.add(ConfigOpt(name='max_logrotate_depth',
                           description=('When log paths are expanded using '
                                        'use_all_logs and they are logrotated '
                                        'log files, this is used to limit the '
                                        'logrotate history in days.'),
                           default_value=7, value_type=int))

    @property
    def name(self):
        return 'search'


class RegisteredOpts(UserDict):
    """ Registers config options. """
    def __init__(self, *optgroups):
        self.optsgroups = []
        data = {}
        for optgroup in optgroups:
            optgroup = optgroup()
            self.optsgroups.append(optgroup)
            a = [name.lower() for name in data]
            b = [name.lower() for name in optgroup]
            if set(a).intersection(b):
                raise NameAlreadyRegisteredError(
                    f"optgroup '{optgroup.name}' contains one or "
                    "more names that have already been registered")

            data.update(optgroup)

        super().__init__(data)

    def __setitem__(self, key, item):
        for group in self.optsgroups:
            if key in group.opts:
                item = group.opts[key].value_type(item)
                break
        else:
            raise KeyError(
                f"config option '{key}' not found in any optgroup")

        self.data[key] = item


class ConfigMeta(abc.ABCMeta):
    """ Metadata used to register config options. """
    REGISTERED = RegisteredOpts(HotSOSConfigOpts,
                                SearchtoolsConfigOpts)
    CONFIG = copy.deepcopy(REGISTERED)

    def __getattr__(cls, key):
        if key in cls.CONFIG:
            return cls.CONFIG[key]

        raise KeyError(f"fetching unknown config '{key}'.")

    def __setattr__(cls, key, val):
        if key in cls.CONFIG:
            cls.CONFIG[key] = val
            return
        if key not in ['__abstractmethods__', '_abc_impl', 'CONFIG',
                       'REGISTERED']:
            raise KeyError(f"setting unknown config '{key}'.")

        super().__setattr__(key, val)


class HotSOSConfig(metaclass=ConfigMeta):
    """ The main registry of config options.

    Options are automatically registered on module load."""

    @classmethod
    def reset(cls):
        """ Reset all config options to their default values. """
        cls.CONFIG = copy.deepcopy(cls.REGISTERED)

    @classmethod
    def set(cls, **config_options):
        """
        Provides a way to set multiple config options at once.

        @param config_options: a dictionary of one or more config option to
                               set.
        """
        for k, v in config_options.items():
            setattr(cls, k, v)



from hotsos.core.plugins.vault import VaultChecks
from hotsos.core.plugintools import summary_entry


class VaultSummary(VaultChecks):
    """ Implementation of Vault summary. """
    summary_part_index = 0

    @summary_entry('services', 0)
    def summary_services(self):
        if self.systemd.services:
            return self.systemd.summary
        if self.pebble.services:
            return self.pebble.summary

        return None

    @summary_entry('snaps', 1)
    def summary_snaps(self):
        if self.snaps.core:
            return self.snaps.all_formatted

        return None
